# import pytorch_lightning as pl

import torch.utils
from torchaudio.transforms import MelScale
import torch.utils.data
import matplotlib.pyplot as plt
import librosa.display
from voicefixer.vocoder.base import Vocoder
from voicefixer.tools.pytorch_util import *
# from voicefixer.models.restorer.mel_denoiser.model_kqq import UNetResComplex_100Mb
from voicefixer.restorer.model_kqq_bn import UNetResComplex_100Mb
from voicefixer.tools.random_ import *
from voicefixer.tools.wav import *
from voicefixer.tools.modules.fDomainHelper import FDomainHelper

from voicefixer.tools.io import load_json, write_json
from matplotlib import cm

os.environ['KMP_DUPLICATE_LIB_OK']='True'
EPS=1e-8

class BN_GRU(torch.nn.Module):
    def __init__(self,input_dim,hidden_dim,layer=1, bidirectional=False, batchnorm=True, dropout=0.0):
        super(BN_GRU, self).__init__()
        self.batchnorm = batchnorm
        if(batchnorm):self.bn = nn.BatchNorm2d(1)
        self.gru = torch.nn.GRU(input_size=input_dim,
                hidden_size=hidden_dim,
                num_layers=layer,
                bidirectional=bidirectional,
                dropout=dropout,
                batch_first=True)
        self.init_weights()

    def init_weights(self):
        for m in self.modules():
            if type(m) in [nn.GRU, nn.LSTM, nn.RNN]:
                for name, param in m.named_parameters():
                    if 'weight_ih' in name:
                        torch.nn.init.xavier_uniform_(param.data)
                    elif 'weight_hh' in name:
                        torch.nn.init.orthogonal_(param.data)
                    elif 'bias' in name:
                        param.data.fill_(0)

    def forward(self,inputs):
        # (batch, 1, seq, feature)
        if(self.batchnorm):inputs = self.bn(inputs)
        out,_ = self.gru(inputs.squeeze(1))
        return out.unsqueeze(1)

class Generator(nn.Module):
    def __init__(self,n_mel,hidden,channels):
        super(Generator, self).__init__()
        # todo the currently running trail don't have dropout
        self.denoiser = nn.Sequential(
            nn.BatchNorm2d(1),
            nn.Linear(n_mel, n_mel * 2),
            nn.ReLU(inplace=True),
            nn.BatchNorm2d(1),
            nn.Linear(n_mel*2, n_mel * 4),
            nn.Dropout(0.5),
            nn.ReLU(inplace=True),
            BN_GRU(input_dim=n_mel * 4, hidden_dim=n_mel * 2, bidirectional=True, layer=2, batchnorm=True),
            BN_GRU(input_dim=n_mel * 4, hidden_dim=n_mel * 2, bidirectional=True, layer=2, batchnorm=True),

            nn.BatchNorm2d(1),
            nn.ReLU(inplace=True),
            nn.Linear(n_mel * 4, n_mel * 4),
            nn.Dropout(0.5),

            nn.BatchNorm2d(1),
            nn.ReLU(inplace=True),
            nn.Linear(n_mel * 4, n_mel),
            nn.Sigmoid()
        )

        self.unet = UNetResComplex_100Mb(channels=channels)

    def forward(self,sp, mel_orig):
        # Denoising
        noisy = mel_orig.clone()
        clean = self.denoiser(noisy) * noisy
        x = to_log(clean.detach())
        unet_in = torch.cat([to_log(mel_orig),x],dim=1)
        # unet_in = lstm_out
        unet_out = self.unet(unet_in)['mel']
        # masks
        mel = unet_out + x
        # todo mel and addition here are in log scales
        return {'mel': mel, "lstm_out":unet_out, "unet_out":unet_out, "noisy": noisy, "clean": clean}

class VoiceFixer(nn.Module):
    def __init__(self, channels, type_target="vocals", nsrc=1, loss="l1",
                 lr=0.002, gamma=0.9,
                 batchsize=None, frame_length=None,
                 sample_rate=None,
                 warm_up_steps=1000, reduce_lr_steps=15000,
                 # datas
                 check_val_every_n_epoch=5,
                 ):
        super(VoiceFixer, self).__init__()

        if(sample_rate == 44100):
            window_size = 2048
            hop_size = 441
            n_mel = 128
        elif(sample_rate == 24000):
            window_size = 768
            hop_size = 240
            n_mel = 80
        elif(sample_rate == 16000):
            window_size = 512
            hop_size = 160
            n_mel = 80
        else:
            raise ValueError("Error: Sample rate "+str(sample_rate)+" not supported")

        center = True,
        pad_mode = 'reflect'
        window = 'hann'
        freeze_parameters = True

        # self.save_hyperparameters()
        self.nsrc = nsrc
        self.type_target = type_target
        self.channels = channels
        self.lr = lr
        self.generated = None
        self.gamma = gamma
        self.sample_rate = sample_rate
        self.sample_rate = sample_rate
        self.batchsize = batchsize
        self.frame_length = frame_length
        # self.hparams['channels'] = 2

        # self.am = AudioMetrics()
        # self.im = ImgMetrics()

        self.vocoder = Vocoder(sample_rate=44100)

        self.valid = None
        self.fake = None

        self.train_step = 0
        self.val_step = 0
        self.val_result_save_dir = None
        self.val_result_save_dir_step = None
        self.downsample_ratio = 2 ** 6  # This number equals 2^{#encoder_blcoks}
        self.check_val_every_n_epoch = check_val_every_n_epoch

        self.f_helper = FDomainHelper(
            window_size=window_size,
            hop_size=hop_size,
            center=center,
            pad_mode=pad_mode,
            window=window,
            freeze_parameters=freeze_parameters,
        )

        hidden = window_size // 2 + 1

        self.mel = MelScale(n_mels=n_mel, sample_rate=sample_rate, n_stft=hidden)

        # masking
        self.generator = Generator(n_mel,hidden,channels)

        self.lr_lambda = lambda step: self.get_lr_lambda(step,
                                                        gamma = self.gamma,
                                                        warm_up_steps=warm_up_steps,
                                                        reduce_lr_steps=reduce_lr_steps)

        self.lr_lambda_2 = lambda step: self.get_lr_lambda(step,
                                                        gamma = self.gamma,
                                                        warm_up_steps=10,
                                                        reduce_lr_steps=reduce_lr_steps)

        self.mel_weight_44k_128 = torch.tensor([19.40951426, 19.94047336, 20.4859038, 21.04629067,
                                       21.62194148, 22.21335214, 22.8210215, 23.44529231,
                                       24.08660962, 24.74541882, 25.42234287, 26.11770576,
                                       26.83212784, 27.56615283, 28.32007747, 29.0947679,
                                       29.89060111, 30.70832636, 31.54828121, 32.41121487,
                                       33.29780773, 34.20865341, 35.14437675, 36.1056621,
                                       37.09332763, 38.10795802, 39.15039691, 40.22119881,
                                       41.32154931, 42.45172373, 43.61293329, 44.80609379,
                                       46.031602, 47.29070223, 48.58427549, 49.91327905,
                                       51.27863232, 52.68119708, 54.1222372, 55.60274206,
                                       57.12364703, 58.68617876, 60.29148652, 61.94081306,
                                       63.63501986, 65.37562658, 67.16408954, 69.00109084,
                                       70.88850318, 72.82736101, 74.81985537, 76.86654792,
                                       78.96885475, 81.12900906, 83.34840929, 85.62810662,
                                       87.97005418, 90.37689804, 92.84887686, 95.38872881,
                                       97.99777002, 100.67862715, 103.43232942, 106.26140638,
                                       109.16827015, 112.15470471, 115.22184756, 118.37439245,
                                       121.6122689, 124.93877158, 128.35661454, 131.86761321,
                                       135.47417938, 139.18059494, 142.98713744, 146.89771854,
                                       150.91684347, 155.0446638, 159.28614648, 163.64270198,
                                       168.12035831, 172.71749158, 177.44220154, 182.29556933,
                                       187.28286676, 192.40502126, 197.6682721, 203.07516896,
                                       208.63088733, 214.33770931, 220.19910108, 226.22363072,
                                       232.41087124, 238.76803591, 245.30079083, 252.01064464,
                                       258.90261676, 265.98474, 273.26010248, 280.73496362,
                                       288.41440094, 296.30489752, 304.41180337, 312.7377183,
                                       321.28877878, 330.07870237, 339.10812951, 348.38276173,
                                       357.91393924, 367.70513992, 377.76413924, 388.09467408,
                                       398.70920178, 409.61813793, 420.81980127, 432.33215467,
                                       444.16083117, 456.30919947, 468.78589276, 481.61325588,
                                       494.78824596, 508.31969844, 522.2238331, 536.51163441,
                                       551.18859414, 566.26142988, 581.75006061, 597.66210737]) / 19.40951426
        self.mel_weight_44k_128 = self.mel_weight_44k_128[None, None, None, ...]


        self.g_loss_weight = 0.01
        self.d_loss_weight = 1

    def get_vocoder(self):
        return self.vocoder

    def get_f_helper(self):
        return self.f_helper

    def get_lr_lambda(self,step, gamma, warm_up_steps, reduce_lr_steps):
        r"""Get lr_lambda for LambdaLR. E.g.,

        .. code-block: python
            lr_lambda = lambda step: get_lr_lambda(step, warm_up_steps=1000, reduce_lr_steps=10000)

            from torch.optim.lr_scheduler import LambdaLR
            LambdaLR(optimizer, lr_lambda)
        """
        if step <= warm_up_steps:
            return step / warm_up_steps
        else:
            return gamma ** (step // reduce_lr_steps)

    def init_weights(self, module: nn.Module):
        for m in module.modules():
            if type(m) in [nn.GRU, nn.LSTM, nn.RNN]:
                for name, param in m.named_parameters():
                    if 'weight_ih' in name:
                        torch.nn.init.xavier_uniform_(param.data)
                    elif 'weight_hh' in name:
                        torch.nn.init.orthogonal_(param.data)
                    elif 'bias' in name:
                        param.data.fill_(0)

    def pre(self, input):
        sp, _, _ = self.f_helper.wav_to_spectrogram_phase(input)
        mel_orig = self.mel(sp.permute(0,1,3,2)).permute(0,1,3,2)
        return sp, mel_orig

    def forward(self, sp, mel_orig):
        """
        Args:
          input: (batch_size, channels_num, segment_samples)

        Outputs:
          output_dict: {
            'wav': (batch_size, channels_num, segment_samples),
            'sp': (batch_size, channels_num, time_steps, freq_bins)}
        """
        return self.generator(sp, mel_orig)

    def configure_optimizers(self):
        optimizer_g = torch.optim.Adam([{'params': self.generator.parameters()}],
                                       lr=self.lr, amsgrad=True, betas=(0.5, 0.999))
        optimizer_d = torch.optim.Adam([{'params': self.discriminator.parameters()}],
                                       lr=self.lr, amsgrad=True,
                                       betas=(0.5, 0.999))

        scheduler_g = {
            'scheduler': torch.optim.lr_scheduler.LambdaLR(optimizer_g, self.lr_lambda),
            'interval': 'step',
            'frequency': 1,
        }
        scheduler_d = {
            'scheduler': torch.optim.lr_scheduler.LambdaLR(optimizer_d, self.lr_lambda),
            'interval': 'step',
            'frequency': 1,
        }
        return [optimizer_g, optimizer_d ], [scheduler_g, scheduler_d]

    def preprocess(self, batch, train=False, cutoff=None):
        if(train):
            vocal = batch[self.type_target] # final target
            noise = batch['noise_LR'] # augmented low resolution audio with noise
            augLR = batch[self.type_target+'_aug_LR'] # # augment low resolution audio
            LR = batch[self.type_target+'_LR']
            # embed()
            vocal, LR, augLR, noise = vocal.float().permute(0, 2, 1), LR.float().permute(0, 2, 1), augLR.float().permute(0, 2, 1), noise.float().permute(0, 2, 1)
            # LR, noise = self.add_random_noise(LR, noise)
            snr, scale = [],[]
            for i in range(vocal.size()[0]):
                vocal[i,...], LR[i,...], augLR[i,...], noise[i,...], _snr, _scale = add_noise_and_scale_with_HQ_with_Aug(vocal[i,...],LR[i,...], augLR[i,...], noise[i,...], snr_l=-5,snr_h=45, scale_lower=0.6, scale_upper=1.0)
                snr.append(_snr), scale.append(_scale)
            # vocal, LR = self.amp_to_original_f(vocal, LR)
            # noise = (noise * 0.0) + 1e-8 # todo
            return vocal, augLR, LR,  noise + augLR
        else:
            if(cutoff is None):
                LR_noisy = batch["noisy"]
                LR = batch["vocals"]
                vocals = batch["vocals"]
                vocals, LR, LR_noisy = vocals.float().permute(0, 2, 1), LR.float().permute(0, 2, 1), LR_noisy.float().permute(0, 2, 1)
                return vocals, LR, LR_noisy, batch['fname'][0]
            else:
                LR_noisy = batch["noisy"+"LR"+"_"+str(cutoff)]
                LR = batch["vocals" + "LR" + "_" + str(cutoff)]
                vocals = batch["vocals"]
                vocals, LR, LR_noisy = vocals.float().permute(0, 2, 1), LR.float().permute(0, 2, 1), LR_noisy.float().permute(0, 2, 1)
                return vocals, LR, LR_noisy, batch['fname'][0]


    def training_step(self, batch, batch_nb, optimizer_idx):
        # dict_keys(['vocals', 'vocals_aug', 'vocals_augLR', 'noise'])
        config = load_json("temp_path.json")
        if("g_loss_weight" not in config.keys()):
            config['g_loss_weight'] = self.g_loss_weight
            config['d_loss_weight'] = self.d_loss_weight
            write_json(config,"temp_path.json")
        elif(config['g_loss_weight'] != self.g_loss_weight or config['d_loss_weight'] != self.d_loss_weight):
            print("Update d_loss weight, from", self.d_loss_weight, "to",config['d_loss_weight'])
            print("Update g_loss weight, from", self.g_loss_weight, "to",config['g_loss_weight'])
            self.g_loss_weight = config['g_loss_weight']
            self.d_loss_weight = config['d_loss_weight']

        if (optimizer_idx == 0):
            self.vocal, self.augLR, _, self.LR_noisy = self.preprocess(batch, train=True)

            for i in range(self.vocal.size()[0]):
                save_wave(tensor2numpy(self.vocal[i, ...]), str(i) + "vocal" + ".wav", sample_rate=44100)
                save_wave(tensor2numpy(self.LR_noisy[i, ...]), str(i) + "LR_noisy" + ".wav", sample_rate=44100)

            # all_mel_e2e in non-log scale
            _, self.mel_target = self.pre(self.vocal)
            self.sp_LR_target, self.mel_LR_target = self.pre(self.augLR)
            self.sp_LR_target_noisy, self.mel_LR_target_noisy = self.pre(self.LR_noisy)

            if (self.valid is None or self.valid.size()[0] != self.mel_target.size()[0]):
                self.valid = torch.ones(self.mel_target.size()[0], 1, self.mel_target.size()[2], 1)
                self.valid = self.valid.type_as(self.mel_target)
            if (self.fake is None or self.fake.size()[0] != self.mel_target.size()[0]):
                self.fake = torch.zeros(self.mel_target.size()[0], 1, self.mel_target.size()[2], 1)
                self.fake = self.fake.type_as(self.mel_target)

            self.generated = self(self.sp_LR_target_noisy, self.mel_LR_target_noisy)

            denoise_loss = self.l1loss(self.generated['clean'], self.mel_LR_target)
            targ_loss = self.l1loss(self.generated['mel'], to_log(self.mel_target))

            self.log("targ-l", targ_loss, on_step=True, on_epoch=False, logger=True, sync_dist=True, prog_bar=True)
            self.log("noise-l", denoise_loss, on_step=True, on_epoch=False, logger=True, sync_dist=True, prog_bar=True)

            loss = targ_loss + denoise_loss

            if(self.train_step >= 18000):
                g_loss = self.bce_loss(self.discriminator(self.generated['mel']), self.valid)
                self.log("g_l", g_loss, on_step=True, on_epoch=False, logger=True, sync_dist=True, prog_bar=True)
                # print("g_loss", g_loss)
                all_loss = loss + self.g_loss_weight * g_loss
                self.log("all_loss", all_loss, on_step=True, on_epoch=True, logger=True, sync_dist=True)
            else:
                all_loss = loss
            self.train_step += 0.5
            return {"loss": all_loss}

        elif(optimizer_idx == 1):
            if(self.train_step >= 16000):
                self.generated = self(self.sp_LR_target_noisy, self.mel_LR_target_noisy)
                self.train_step += 0.5
                real_loss = self.bce_loss(self.discriminator(to_log(self.mel_target)),self.valid)
                self.log("r_l", real_loss, on_step=True, on_epoch=False, logger=True, sync_dist=True, prog_bar=True)
                fake_loss = self.bce_loss(self.discriminator(self.generated['mel'].detach()), self.fake)
                self.log("d_l", fake_loss, on_step=True, on_epoch=False, logger=True, sync_dist=True, prog_bar=True)
                d_loss = self.d_loss_weight * (real_loss+fake_loss) / 2
                self.log("discriminator_loss", d_loss, on_step=True, on_epoch=True, logger=True, sync_dist=True)
                return {"loss": d_loss}

    def draw_and_save(self, mel: torch.Tensor, path, clip_max=None, clip_min=None, needlog=True):
        plt.figure(figsize=(15,5))
        if(clip_min is None):
            clip_max,clip_min = self.clip(mel)
        mel = np.transpose(tensor2numpy(mel)[0,0,...],(1,0))
        # assert np.sum(mel < 0) == 0, str(np.sum(mel < 0)) + str(np.sum(mel < 0))

        if(needlog):
            assert np.sum(mel < 0) == 0, str(np.sum(mel < 0))+"-"+path
            mel_log = np.log10(mel+EPS)
        else:
            mel_log = mel

        # plt.imshow(mel)
        librosa.display.specshow(mel_log, sr=44100,x_axis='frames',y_axis='mel',cmap=cm.jet, vmax=clip_max, vmin=clip_min)
        plt.colorbar()
        plt.savefig(path)
        plt.close()

    def clip(self,*args):
        val_max, val_min = [],[]
        for each in args:
            val_max.append(torch.max(each))
            val_min.append(torch.min(each))
        return max(val_max), min(val_min)




#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   test.py
@Contact :   haoheliu@gmail.com
@License :   (C)Copyright 2020-2100

@Modify Time      @Author    @Version    @Desciption
------------      -------    --------    -----------
9/14/21 11:02 AM   Haohe Liu      1.0         None
'''

import git
import os
import sys
import librosa
import numpy as np
import torch

git_root = git.Repo("", search_parent_directories=True).git.rev_parse("--show-toplevel")
sys.path.append(git_root)
from voicefixer import VoiceFixer, Vocoder

os.makedirs(os.path.join(git_root,"test/utterance/output"),exist_ok=True)

def check(fname):
    """
    check if the output is normal
    """
    output = os.path.join(git_root,"test/utterance/output",fname)
    target = os.path.join(git_root, "test/utterance/target", fname)
    output, _ = librosa.load(output,sr=44100)
    target, _ = librosa.load(target, sr=44100)
    assert np.mean(np.abs(output-target)) < 0.01

# TEST VOICEFIXER
## Initialize a voicefixer
print("Initializing VoiceFixer...")
voicefixer = VoiceFixer()
# Mode 0: Original Model (suggested by default)
# Mode 1: Add preprocessing module (remove higher frequency)
# Mode 2: Train mode (might work sometimes on seriously degraded real speech)
for mode in [0,1,2]:
    print("Test voicefixer mode",mode,end=", ")
    voicefixer.restore(input=os.path.join(git_root,"test/utterance/original/original.flac"), # low quality .wav/.flac file
                       output=os.path.join(git_root,"test/utterance/output/output_mode_"+str(mode)+".flac"), # save file path
                       cuda=False, # GPU acceleration
                       mode=mode)
    if (mode != 2):
        check("output_mode_" + str(mode) + ".flac")

    if(torch.cuda.is_available()):
        voicefixer.restore(input=os.path.join(git_root, "test/utterance/original/original.flac"),
                           # low quality .wav/.flac file
                           output=os.path.join(git_root, "test/utterance/output/output_mode_" + str(mode) + ".flac"),
                           # save file path
                           cuda=True,  # GPU acceleration
                           mode=mode)
    if(mode != 2):
        check("output_mode_"+str(mode)+".flac")
    print("Pass")

# TEST VOCODER
## Initialize a vocoder
print("Initializing 44.1kHz speech vocoder...")
vocoder = Vocoder(sample_rate=44100)

### read wave (fpath) -> mel spectrogram -> vocoder -> wave -> save wave (out_path)
print("Test vocoder using groundtruth mel spectrogram...")
vocoder.oracle(fpath=os.path.join(git_root,"test/utterance/original/p360_001_mic1.flac"),
               out_path=os.path.join(git_root,"test/utterance/output/oracle.flac"),
               cuda=False) # GPU acceleration

check("oracle.flac")

if(torch.cuda.is_available()):
    vocoder.oracle(fpath=os.path.join(git_root, "test/utterance/original/p360_001_mic1.flac"),
                   out_path=os.path.join(git_root, "test/utterance/output/oracle.flac"),
                   cuda=True)  # GPU acceleration
# Another interface
# vocoder.forward(mel=mel)
check("oracle.flac")

print("Pass")




import wave
import os
import numpy as np
import scipy.signal as signal
import soundfile as sf
import librosa

def save_wave(frames: np.ndarray, fname, sample_rate=44100):
    shape = list(frames.shape)
    if(len(shape) == 1):
        frames = frames[...,None]
    in_samples, in_channels = shape[-2], shape[-1]
    if(in_channels >= 3):
        if(len(shape) == 2):
            frames = np.transpose(frames,(1,0))
        elif(len(shape) == 3):
            frames = np.transpose(frames, (0,2,1))
        msg = "Warning: Save audio with "+str(in_channels) +" channels, save permute audio with shape "+str(list(frames.shape))+" please check if it's correct."
        # print(msg)
    if(np.max(frames) <= 1 and frames.dtype == np.float32 or frames.dtype == np.float16 or frames.dtype == np.float64):
        frames *= 2**15
    frames = frames.astype(np.short)
    if (len(frames.shape) >= 3):
        frames = frames[0,...]
    sf.write(fname,frames,samplerate=sample_rate)

def constrain_length(chunk, length):
    frames_length = chunk.shape[0]
    if(frames_length == length):
        return chunk
    elif(frames_length < length):
        return np.pad(chunk, ((0, int(length - frames_length)),(0,0)), 'constant')
    else:
        return chunk[:length,...]

def random_chunk_wav_file(fname, chunk_length):
    '''
    fname: path to wav file
    chunk_length: frame length in seconds
    '''
    with wave.open(fname) as f:
        params = f.getparams()
        duration = params[3] / params[2]
        sample_rate = params[2]
        sample_length = params[3]
        if(duration < chunk_length or abs(duration-chunk_length)<1e-4):
            frames = read_wave(fname,sample_rate)
            return frames, duration, sample_rate # [-1,1]
        else:
            # Random trunk
            random_starts = np.random.randint(0,sample_length-sample_rate*chunk_length)
            random_end = (random_starts+sample_rate*chunk_length)
            random_starts, random_end = random_starts/sample_rate,random_end/sample_rate
            random_starts, random_end = random_starts/duration,random_end/duration
            frames = read_wave(fname,sample_rate,portion_start=random_starts,portion_end=random_end)
            frames = constrain_length(frames,length=int(chunk_length*sample_rate))
            return frames, chunk_length, sample_rate

def random_chunk_wav_file_v2(fname, chunk_length, random_starts=None, random_end=None):
    '''
    fname: path to wav file
    chunk_length: frame length in seconds
    '''
    with wave.open(fname) as f:
        params = f.getparams()
        duration = params[3] / params[2]
        sample_rate = params[2]
        sample_length = params[3]
        if(duration < chunk_length or abs(duration-chunk_length)<1e-4):
            frames = read_wave(fname,sample_rate)
            return frames, duration, sample_rate # [-1,1]
        else:
            # Random trunk
            if(random_starts is None and random_end is None):
                random_starts = np.random.randint(0,sample_length-sample_rate*chunk_length)
                random_end = (random_starts+sample_rate*chunk_length)
                random_starts, random_end = random_starts/sample_rate,random_end/sample_rate
                random_starts, random_end = random_starts/duration,random_end/duration
            frames = read_wave(fname,sample_rate,portion_start=random_starts,portion_end=random_end)
            frames = constrain_length(frames,length=int(chunk_length*sample_rate))
            return frames, chunk_length, sample_rate, random_starts, random_end

def read_wave(fname,
              sample_rate,
              portion_start = 0,
              portion_end = 1,
              ): # Whether you want raw bytes
    """
    :param fname: wav file path
    :param sample_rate:
    :param portion_start:
    :param portion_end:
    :return: [sample, channels]
    """
    # sr = get_sample_rate(fname)
    # if(sr != sample_rate):
    #     print("Warning: Sample rate not match, may lead to unexpected behavior.")
    if(portion_end > 1 and portion_end < 1.1):
        portion_end = 1
    if(portion_end != 1):
        duration = get_duration(fname)
        wav, _ = librosa.load(fname,sr=sample_rate,
                              offset=portion_start*duration,
                              duration=(portion_end-portion_start)*duration,
                              mono=False)
    else:
        wav, _ = librosa.load(fname, sr=sample_rate, mono=False)
    if(len(list(wav.shape)) == 1):
        wav = wav[...,None]
    else:
        wav = np.transpose(wav,(1,0))
    return wav



def get_channels_sampwidth_and_sample_rate(fname):
    with wave.open(fname) as f:
        params = f.getparams()
    return params[0],params[1],params[2] # == (2,2,44100),(params[0],params[1],params[2])

def get_channels(fname):
    with wave.open(fname) as f:
        params = f.getparams()
    return params[0]

def get_sample_rate(fname):
    with wave.open(fname) as f:
        params = f.getparams()
    return params[2]

def get_duration(fname):
    with wave.open(fname) as f:
        params = f.getparams()
    return params[3]/params[2]

def get_framesLength(fname):
    with wave.open(fname) as f:
        params = f.getparams()
    return params[3]


def restore_wave(zxx):
    _,w = signal.istft(zxx)
    return w

def calculate_total_times(dir):
    total = 0
    for each in os.listdir(dir):
        fname = os.path.join(dir,each)
        try:
            duration = get_duration(fname)
        except:
            print(fname)
        total += duration
    return total

def filter(pth):
    global dic
    temp = []
    for each in os.listdir(pth):
        temp.append(os.path.join(pth,each))
    for each in temp:
        sr = get_sample_rate(each)
        if(sr not in dic.keys()):
            dic[sr] = []
        dic[sr].append(each)
    for each in dic[16000]:
        # print(each)
        pass
    print(dic.keys())
    for each in list(dic.keys()):
        print(each,len(dic[each]))


if __name__ == "__main__":
    path = "/Users/admin/Desktop/p376_025.wav"
    stereo = "/Users/admin/Desktop/vocals.wav"
    path_16 = "/Users/admin/Desktop/SI869.WAV.wav"
    import time
    start = time.time()
    for i in range(1000):
        frames, duration, sample_rate = random_chunk_wav_file(stereo, chunk_length=3.0)
        print(frames.shape,np.max(frames))
        save_wave(frames, "stero.wav", sample_rate=44100)
        frames, duration, sample_rate = random_chunk_wav_file(path, chunk_length=3.0)
        print(frames.shape,np.max(frames))
        save_wave(frames, "mono.wav", sample_rate=44100)
        frames, duration, sample_rate = random_chunk_wav_file(path_16, chunk_length=3.0)
        print(frames.shape,np.max(frames))
        save_wave(frames, "16.wav", sample_rate=16000)
    print(time.time()-start)
    # frames = read_wave(stereo,sample_rate=44100)
    print(frames.shape)

    print(frames)


import os
import sys
import torch
import torch.nn as nn
import numpy as np
import scipy.io.wavfile

class PQMF(nn.Module):
    def __init__(self, N, M, file_path="utils/pqmf_hk_4_64.dat"):
        super().__init__()
        self.N = N #nsubband
        self.M = M #nfilter
        self.ana_conv_filter = nn.Conv1d(1, out_channels=N, kernel_size=M, stride=N, bias=False)
        data=np.reshape(np.fromfile(file_path, dtype=np.float32), (N, M))
        data=np.flipud(data.T).T
        gk=data.copy()
        data=np.reshape(data, (N, 1, M)).copy()
        dict_new = self.ana_conv_filter.state_dict().copy()
        dict_new['weight'] = torch.from_numpy(data)
        self.ana_pad = nn.ConstantPad1d((M-N,0), 0)
        self.ana_conv_filter.load_state_dict(dict_new)

        self.syn_pad = nn.ConstantPad1d((0, M//N-1), 0)
        self.syn_conv_filter = nn.Conv1d(N, out_channels=N, kernel_size=M//N, stride=1, bias=False)
        gk=np.transpose(np.reshape(gk,(4, 16, 4)), (1, 0, 2))*N
        gk=np.transpose(gk[::-1,:,:], (2, 1, 0)).copy()
        dict_new = self.syn_conv_filter.state_dict().copy()
        dict_new['weight'] = torch.from_numpy(gk)
        self.syn_conv_filter.load_state_dict(dict_new)

        for param in self.parameters():
            param.requires_grad = False

    def analysis(self, inputs):
        return self.ana_conv_filter(self.ana_pad(inputs))
    def synthesis(self, inputs):
        return self.syn_conv_filter(self.syn_pad(inputs))
    def forward(self, inputs):
        return self.ana_conv_filter(self.ana_pad(inputs))

if __name__ == "__main__":
    a=PQMF(4, 64)
    #x = np.load('data/train/audio/010000.npy')
    x = np.zeros([8, 24000], np.float32)
    x=np.reshape(x, (8, 1, -1))
    x = torch.from_numpy(x)
    b=a.analysis(x)
    c=a.synthesis(b)
    print(x.shape, b.shape, c.shape)
    b=(b * 32768).numpy()
    b=np.reshape(np.transpose(b, (0, 2, 1)), (-1, 1)).astype(np.int16)
    #b.tofile('1.pcm')
    #np.reshape(np.transpose(c.numpy()*32768, (0, 2, 1)), (-1,1)).astype(np.int16).tofile('2.pcm')



import json
# from scipy.io import loadmat
from pathlib import Path
import yaml
import pickle

def read_list(fname):
    result = []
    with open(fname, "r") as f:
        for each in f.readlines():
            each = each.strip('\n')
            result.append(each)
    return result

def write_list(list, fname):
    with open(fname,'w') as f:
        for word in list:
            f.write(word)
            f.write('\n')

def write_json(my_dict, fname):
    # print("Save json file at "+fname)
    json_str = json.dumps(my_dict)
    with open(fname, 'w') as json_file:
        json_file.write(json_str)

def load_json(fname):
    with open(fname,'r') as f:
        data = json.load(f)
        return data


# def load_mat2numpy(fname=""):
#     '''
#     Args:
#         fname: pth to mat
#         type:
#     Returns: dic object
#     '''
#     if (len(fname) == 0):
#         return None
#     else:
#         return loadmat(fname)

def save_pickle(obj,fname):
    # print("Save pickle at "+fname)
    with open(fname,'wb') as f:
        pickle.dump(obj,f)

def load_pickle(fname):
    # print("Load pickle at "+fname)
    with open(fname,'rb') as f:
        res = pickle.load(f)
    return res

def read_yaml(fname):
    fname = Path(fname)
    with fname.open('rt') as handle:
        return yaml.load(handle, Loader=yaml.FullLoader)


def write_yaml(content, fname):
    fname = Path(fname)
    with fname.open('wt') as handle:
        yaml.dump(content, handle)

# if __name__ =="__main__":
#     a = {"sdf":float(1.334234)}
#     write_json(a,"temp.json")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   __init__.py.py    
@Contact :   haoheliu@gmail.com
@License :   (C)Copyright 2020-2100

@Modify Time      @Author    @Version    @Desciption
------------      -------    --------    -----------
9/14/21 12:31 AM   Haohe Liu      1.0         None
'''

import os

meta = {
    "voicefixer_fe": {
        "path": os.path.join(os.path.expanduser('~'), ".cache/voicefixer/analysis_module/checkpoints/epoch=15_trimed_bn.ckpt"),
        "url": "https://zenodo.org/record/5464142/files/epoch%3D15_trimed.ckpt?download=1",
    },
}

if (not os.path.exists(meta["voicefixer_fe"]['path'])):
    os.makedirs(os.path.dirname(meta["voicefixer_fe"]['path']), exist_ok=True)
    print("Downloading the main structure of voicefixer")
    cmd = "wget https://zenodo.org/record/5469951/files/epoch%3D15_trimed_bn.ckpt?download=1 -O " + \
          meta["voicefixer_fe"]['path']
    os.system(cmd)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   __init__.py.py    
@Contact :   haoheliu@gmail.com
@License :   (C)Copyright 2020-2100

@Modify Time      @Author    @Version    @Desciption
------------      -------    --------    -----------
9/14/21 1:00 AM   Haohe Liu      1.0         None
'''

import os
from voicefixer.vocoder.config import Config

if (not os.path.exists(Config.ckpt)):
    os.makedirs(os.path.dirname(Config.ckpt), exist_ok=True)
    print("Downloading the weight of neural vocoder: TFGAN")
    cmd = "wget https://zenodo.org/record/5469951/files/model.ckpt-1490000_trimed.pt?download=1 -O " + Config.ckpt
    os.system(cmd)



from voicefixer.vocoder.config import Config
from voicefixer.tools.pytorch_util import try_tensor_cuda,check_cuda_availability
import torch
import librosa
import numpy as np

def tr_normalize(S):
    if Config.allow_clipping_in_normalization:
        if Config.symmetric_mels:
            return torch.clip((2 * Config.max_abs_value) * ((S - Config.min_db) / (-Config.min_db)) - Config.max_abs_value, -Config.max_abs_value,
                           Config.max_abs_value)
        else:
            return torch.clip(
                Config.max_abs_value * ((S - Config.min_db) /
                                         (-Config.min_db)), 0,
                Config.max_abs_value)

    assert S.max() <= 0 and S.min() - Config.min_db >= 0
    if Config.symmetric_mels:
        return ((2 * Config.max_abs_value) * ((S - Config.min_db) /
                                               (-Config.min_db)) -
                Config.max_abs_value)
    else:
        return (Config.max_abs_value * ((S - Config.min_db) /
                                         (-Config.min_db)))

def tr_amp_to_db(x):
    min_level = torch.exp(Config.min_level_db / 20 * torch.log(torch.tensor(10.0)))
    min_level = min_level.type_as(x)
    return 20 * torch.log10(torch.maximum(min_level, x))

def normalize(S):
    if Config.allow_clipping_in_normalization:
        if Config.symmetric_mels:
            return np.clip((2 * Config.max_abs_value) * ((S - Config.min_db) /
                                                          (-Config.min_db)) -
                           Config.max_abs_value, -Config.max_abs_value,
                           Config.max_abs_value)
        else:
            return np.clip(
                Config.max_abs_value * ((S - Config.min_db) /
                                         (-Config.min_db)), 0,
                Config.max_abs_value)

    assert S.max() <= 0 and S.min() - Config.min_db >= 0
    if Config.symmetric_mels:
        return ((2 * Config.max_abs_value) * ((S - Config.min_db) /
                                               (-Config.min_db)) -
                Config.max_abs_value)
    else:
        return (Config.max_abs_value * ((S - Config.min_db) /
                                         (-Config.min_db)))

def amp_to_db(x):
    min_level = np.exp(Config.min_level_db / 20 * np.log(10))
    return 20 * np.log10(np.maximum(min_level, x))

def tr_pre(npy):
    # conditions = torch.FloatTensor(npy).type_as(npy) # to(device)
    conditions = npy.transpose(1, 2)
    l = conditions.size(-1)
    pad_tail = l % 2 + 4
    zeros = torch.zeros([conditions.size()[0], Config.num_mels, pad_tail]).type_as(conditions) + -4.0
    return torch.cat([conditions, zeros], dim=-1)

def pre(npy):
    conditions = npy
    ## padding tail
    if(type(conditions) == np.ndarray):
        conditions = torch.FloatTensor(conditions).unsqueeze(0)
    else:
        conditions = torch.FloatTensor(conditions.float()).unsqueeze(0)
    conditions = conditions.transpose(1, 2)
    l = conditions.size(-1)
    pad_tail = l % 2 + 4
    zeros = torch.zeros([1, Config.num_mels, pad_tail]) + -4.0
    return torch.cat([conditions, zeros], dim=-1)

def load_try(state, model):
    model_dict = model.state_dict()
    try:
        model_dict.update(state)
        model.load_state_dict(model_dict)
    except RuntimeError as e:
        print(str(e))
        model_dict = model.state_dict()
        for k, v in state.items():
            model_dict[k] = v
            model.load_state_dict(model_dict)

def load_checkpoint(checkpoint_path, device):
    checkpoint = torch.load(checkpoint_path, map_location=device)
    return checkpoint

def build_mel_basis():
    return librosa.filters.mel(
        Config.sample_rate,
        Config.n_fft,
        htk=True,
        n_mels=Config.num_mels,
        fmin=0,
        fmax=int(Config.sample_rate // 2))


def linear_to_mel(spectogram):
    _mel_basis = build_mel_basis()
    return np.dot(_mel_basis, spectogram)


if __name__ == "__main__":
    data = torch.randn((3,5,100))
    b = normalize(amp_to_db(data.numpy()))
    a = tr_normalize(tr_amp_to_db(data)).numpy()
    print(a-b)
